/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

/* This is a copy of v5/src/lib/sfgpxe/dyncfg_accessors.c.
 * When possible, changes should be done in the v5 repo and backported here.
 *
 * Differences with the original file should be noted with a 'DIFF' marker, and
 * a comment explaining why the difference is kept local.
 */

/* DIFF: Update paths for the gpxe build system. */
#include <errno.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>
#include "dyncfg_accessors.h"
#include "tlv_layout.h"
#include "tlv_partition.h"
#include "tlv_platform.h"
#include "pci_ids.h"
#include "endian_base.h"
#include <gpxe/sf_main.h>

#define TLV_ITEM_NON_DATA_SIZE (TLV_STRUCT_TAG_SIZE + TLV_STRUCT_LENGTH_SIZE)

unsigned int get_port_count(uint32_t port_mode)
{
	switch(port_mode) {
	case TLV_PORT_MODE_DEFAULT:
		/* The number of ports cannot be determined if port_mode is
		 * DEFAULT, so assume that there are two.
		 */
		return 2;
	case TLV_PORT_MODE_10G:
	case TLV_PORT_MODE_40G:
	case TLV_PORT_MODE_10G_10G_10G_10G_Q2:
		return 1;
	case TLV_PORT_MODE_10G_10G_10G_10G:
	case TLV_PORT_MODE_10G_10G_10G_10G_Q:
		return 4;
	case TLV_PORT_MODE_40G_40G:
	case TLV_PORT_MODE_10G_10G:
	case TLV_PORT_MODE_1x2_1x2:
	case TLV_PORT_MODE_2x2_NA:
		return 2;
	case TLV_PORT_MODE_40G_10G_10G:
	case TLV_PORT_MODE_10G_10G_40G:
		return 3;
	default:
		DBG("%s unknown port mode %x\n", __FUNCTION__, port_mode);
		return 0;
	}
}

static unsigned get_stride_count(uint32_t port_mode)
{
	unsigned stride;

	switch(port_mode) {
	case TLV_PORT_MODE_DEFAULT:
	case TLV_PORT_MODE_10G:
	case TLV_PORT_MODE_40G:
	case TLV_PORT_MODE_10G_10G:
	case TLV_PORT_MODE_10G_10G_10G_10G:
	case TLV_PORT_MODE_10G_10G_10G_10G_Q:
	case TLV_PORT_MODE_10G_10G_10G_10G_Q2:
	case TLV_PORT_MODE_1x2_1x2:
	case TLV_PORT_MODE_2x2_NA:
		stride = 1;
		break;
	case TLV_PORT_MODE_40G_40G:
	case TLV_PORT_MODE_40G_10G_10G:
	case TLV_PORT_MODE_10G_10G_40G:
		stride = 2;
		break;
	default:
		stride = 0;
		DBG("%s unknown port mode %x\n", __FUNCTION__, port_mode);
	}
	return stride;
}

void tlv_per_pf_pcie_config_initialise(struct tlv_per_pf_pcie_config *s,
                                       size_t unused __maybe_unused)
{
  /* The firmware looks for 0xFF uninitialised marker values and will use
   * appropriate default values.
   * It does not do that for the pf_bar0_aperture,
   * pf_bar2_aperture, and vf_bar0_aperture
   * members, which are not per-PF settings, and should actually not have
   * been in this structure, but are left for layout compatibility.
   * Adding defaults for vectors per pf, vectors per vfs and
   * vf-count. This is needed in order for check_resources() to work
   * correctly.
   */
  memset(s, 0xFF, sizeof(*s));
  s->pf_bar0_aperture = DEFAULT_PF_BAR0_APERTURE_HUNT;
  s->pf_bar2_aperture = DEFAULT_PF_BAR2_APERTURE_HUNT;
  s->vf_bar0_aperture = DEFAULT_VF_BAR0_APERTURE_HUNT;
  s->vectors_per_pf = DYNCFG_SETTING_DEFAULT_MSIX_LIMIT;
  s->vectors_per_vf = DYNCFG_SETTING_DEFAULT_VF_MSIX_LIMIT;
  s->vfs_total = DYNCFG_SETTING_DEFAULT_VF_COUNT;
}

static int min(int a, int b) { return a < b ? a : b; }

static uint8_t calculate_vf_bar0_aperture(unsigned n_vfs, unsigned port_count)
{
	unsigned vf_count_per_port = n_vfs / port_count;
	unsigned round_up = (( (vf_count_per_port & (vf_count_per_port - 1)) == 0) ? 0 : 1);
	unsigned log_vfs = 0;
	unsigned max_vi_mem_per_vf;

	/* log the VFs per port, rounding up */
	while(vf_count_per_port > 1) {
		vf_count_per_port >>= 1;
		++log_vfs;
	}
	log_vfs += round_up;

	/* max mem per VF is log2(MAX_VI_MEM_PER_PORT * GENEROSITY_FACTOR / num_vfs).
	 * values are log2 so add and subtract instead of multiply and divide */
	max_vi_mem_per_vf = MAX_VI_MEMORY_PER_PORT_LOG2 + GENEROSITY_FACTOR_LOG2 - log_vfs;
	return min(max_vi_mem_per_vf, MAX_VI_MEMORY_PER_PORT_LOG2);
}

int update_vf_aperture(unsigned char *dyncfg_buf, int dyncfg_buf_size,
		       struct sfgpxe_dev_info *dev_info, uint32_t port_mode)
{
	int rc;
	unsigned n_pfs, n_vfs = 0, i, port_count;
	nvram_partition_t partition;
	tlv_cursor_t *cursor = &partition.tlv_cursor;
	struct tlv_per_pf_pcie_config pf_config;
	struct tlv_pcie_config *pcie_config;
	uint8_t vf_aperture;

	port_count = get_port_count(port_mode);
	if (port_count == 0)
		return -EINVAL;

	rc = tlv_init_partition_from_buffer(&partition, dyncfg_buf,
					    dyncfg_buf_size);
	if (rc)
		return rc;

	rc = tlv_find(cursor, TLV_TAG_GLOBAL_PCIE_CONFIG);
	if (rc)
		return rc;

	pcie_config = (struct tlv_pcie_config*)cursor->current;
	n_pfs = le16_to_host(pcie_config->max_pf_number) + 1;

	for (i = 0; i < n_pfs; ++i) {
		rc = tlv_find(cursor, TLV_TAG_PF_PCIE_CONFIG(i));
		if (rc == TLV_OK)
			memcpy(&pf_config,
			       (struct tlv_per_pf_pcie_config*)cursor->current,
			       sizeof(pf_config));
		else if (rc == TLV_NOT_FOUND)
			tlv_per_pf_pcie_config_initialise(&pf_config,
							  sizeof(pf_config));
		else
			return rc;

		n_vfs += pf_config.vfs_total;
	}

	vf_aperture = calculate_vf_bar0_aperture(n_vfs, port_count);

	return gpxe_dyncfg_set_vf_aper(dyncfg_buf, dyncfg_buf_size, dev_info, vf_aperture);
}

int update_per_pf_pcie_config(unsigned char *dyncfg_buf, int dyncfg_buf_size,
			      uint32_t pci_device_id, uint32_t port_mode)
{
	int rc;
	unsigned n_pfs, stride, i, port_count, next_free_rid;
	nvram_partition_t partition;
	tlv_cursor_t *cursor = &partition.tlv_cursor;
	struct tlv_pcie_config *pcie_config;
	struct tlv_per_pf_pcie_config pf_config;
	void *pf_config_data = &(pf_config.vfs_total);
	size_t pf_config_data_size = sizeof(pf_config) -
				     offsetof(struct tlv_per_pf_pcie_config,
					      vfs_total);

	port_count = get_port_count(port_mode);
	stride = get_stride_count(port_mode);
	if (!port_count || !stride)
		return -EIO;

	if ((port_mode == TLV_PORT_MODE_DEFAULT) &&
	    (SFU_DEVICE_TYPE(pci_device_id) == SFC_GREENPORT_PF_DEVICE_ID))
		stride = 2;

	rc = tlv_init_partition_from_buffer(&partition, dyncfg_buf,
					    dyncfg_buf_size);
	if (rc)
		return rc;

	rc = tlv_find(cursor, TLV_TAG_GLOBAL_PCIE_CONFIG);
	if (rc)
		return rc;
	pcie_config = (struct tlv_pcie_config*)cursor->current;

	n_pfs = le16_to_host(pcie_config->max_pf_number) + 1;
	next_free_rid = n_pfs;

	for (i = 0; i < n_pfs; ++i) {
		rc = tlv_find(cursor, TLV_TAG_PF_PCIE_CONFIG(i));
		if (rc == TLV_OK)
			memcpy(&pf_config,
			       (struct tlv_per_pf_pcie_config*)cursor->current,
			       sizeof(pf_config));
		else if (rc == TLV_NOT_FOUND)
			tlv_per_pf_pcie_config_initialise(&pf_config,
							  sizeof(pf_config));
		else
			return rc;

		pf_config.port_allocation = (i % port_count) * stride;
		if (pf_config.vfs_total != 0 && pf_config.vfs_total != 0xff) {
			pf_config.vf_base = next_free_rid - i;
			next_free_rid += pf_config.vfs_total;
		} else {
			pf_config.vf_base = 0xff;
		}

		rc = tlv_update_or_append_to_partition(&partition,
						       TLV_TAG_PF_PCIE_CONFIG(i),
						       (const uint8_t*)pf_config_data,
						       pf_config_data_size);
		if (rc)
			return rc;
	}

	for (i = n_pfs; i < MAX_PF_COUNT; ++i) {
		rc = tlv_find(cursor, TLV_TAG_PF_PCIE_CONFIG(i));
		if (rc == TLV_OK)
			tlv_delete_from_partition(cursor);
		else if (rc == TLV_NOT_FOUND)
			continue;
		else
			return -EINVAL;
	}

	return 0;
}

#define DEFINE_PER_PF_PCIE_CONFIG_ACCESSORS(                                   \
    _opt_name, _val_type,                                                      \
    _host_to_val_type, _val_type_to_host,                                      \
    _struct_member)                                                            \
                                                                               \
void gpxe_dyncfg_get_##_opt_name(                                              \
    unsigned char *dyncfg_buf, int dyncfg_buf_size,                            \
    struct sfgpxe_dev_info *dev_info __maybe_unused,                           \
    _val_type *res, _val_type default_val)                                     \
{                                                                              \
  nvram_partition_t partition;                                                 \
  tlv_cursor_t *cursor = &partition.tlv_cursor;                                \
  uint32_t tlv_tag = TLV_TAG_PF_PCIE_CONFIG(dev_info->pf_index);               \
  struct tlv_per_pf_pcie_config *cfg_struct;                                   \
                                                                               \
  if (tlv_init_partition_from_buffer(&partition,                               \
                                     dyncfg_buf, dyncfg_buf_size)) {           \
    goto default_value;                                                        \
  }                                                                            \
                                                                               \
  if (tlv_find(cursor, tlv_tag))                                               \
    goto default_value;                                                        \
                                                                               \
  cfg_struct = (struct tlv_per_pf_pcie_config*)cursor->current;                \
  *res = _val_type_to_host(cfg_struct->_struct_member);                        \
  if (*res == (_val_type)-1)                                                   \
    goto default_value;                                                        \
  return;                                                                      \
                                                                               \
default_value:                                                                 \
  *res = default_val;                                                          \
}                                                                              \
                                                                               \
int gpxe_dyncfg_set_##_opt_name(                                               \
    unsigned char *dyncfg_buf, int dyncfg_buf_size,                            \
    struct sfgpxe_dev_info *dev_info __maybe_unused,                           \
    _val_type val)                                                             \
{                                                                              \
  nvram_partition_t partition;                                                 \
  tlv_cursor_t *cursor = &partition.tlv_cursor;                                \
  uint32_t tlv_tag = TLV_TAG_PF_PCIE_CONFIG(dev_info->pf_index);               \
  struct tlv_per_pf_pcie_config cfg_struct;                                    \
  void *cfg_struct_data = &(cfg_struct.vfs_total);                             \
  size_t cfg_struct_data_size = sizeof(cfg_struct) - TLV_ITEM_NON_DATA_SIZE;   \
  int rc;								       \
                                                                               \
  /* The firmware looks for 0xFF uninitialised marker values and will use      \
   * appropriate default values.                                               \
   * It does not do that for the                                               \
   *    pf_bar0_aperture, pf_bar2_aperture, vf_bar0_aperture,                  \
   * members, which are not per-PF settings, should actually not have been in  \
   * this structure, but are left for layout compatibility.                    \
   */                                                                          \
  memset(&cfg_struct, 0xFF, sizeof(cfg_struct));                               \
  cfg_struct.pf_bar0_aperture = DEFAULT_PF_BAR0_APERTURE_HUNT;                 \
  cfg_struct.pf_bar2_aperture = DEFAULT_PF_BAR2_APERTURE_HUNT;                 \
  cfg_struct.vf_bar0_aperture = DEFAULT_VF_BAR0_APERTURE_HUNT;                 \
                                                                               \
  rc = tlv_init_partition_from_buffer(&partition, dyncfg_buf, dyncfg_buf_size); \
  if (rc) {								       \
    DBGC(dyncfg_buf, "%s: TLV init failed with %x\n", __func__, rc);	       \
    return -EINVAL;                                                            \
  }                                                                            \
                                                                               \
  if (TLV_OK == tlv_find(&partition.tlv_cursor, tlv_tag)) {                    \
    memcpy(&cfg_struct,                                                        \
           (struct tlv_per_pf_pcie_config*)cursor->current,                    \
           sizeof(cfg_struct));                                                \
  }                                                                            \
                                                                               \
  cfg_struct._struct_member = _host_to_val_type(val);                          \
                                                                               \
  rc = tlv_update_or_append_to_partition(                                      \
      &partition,                                                              \
      tlv_tag,                                                                 \
      (const uint8_t*)cfg_struct_data, cfg_struct_data_size);                  \
  if (rc)								       \
    DBGC(dyncfg_buf, "%s: TLV update failed with %x\n", __func__, rc);	       \
  return rc;								       \
}

DEFINE_PER_PF_PCIE_CONFIG_ACCESSORS(msix_limit, uint16_t,
                                    host_to_le16, le16_to_host, vectors_per_pf)
DEFINE_PER_PF_PCIE_CONFIG_ACCESSORS(vf_msix_limit, uint16_t, , , vectors_per_vf)
DEFINE_PER_PF_PCIE_CONFIG_ACCESSORS(vf_count, uint8_t, , , vfs_total)


#define DEFINE_CONFIG_ACCESSORS(_tlv_tag, _config_struct,                      \
                                _config_struct_first_data_member,              \
                                _opt_name, _val_type, _struct_member,          \
                                _host_to_val_type, _val_type_to_host)          \
                                                                               \
void gpxe_dyncfg_get_##_opt_name(                                              \
    unsigned char *dyncfg_buf,                                                 \
    int dyncfg_buf_size,                                                       \
    struct sfgpxe_dev_info *dev_info __maybe_unused,                           \
    _val_type *res, _val_type default_val)                                     \
{                                                                              \
  nvram_partition_t partition;                                                 \
  tlv_cursor_t *cursor = &partition.tlv_cursor;                                \
  struct _config_struct *cfg_struct;                                           \
                                                                               \
  if (tlv_init_partition_from_buffer(&partition,                               \
                                     dyncfg_buf, dyncfg_buf_size)) {           \
    goto default_value;                                                        \
  }                                                                            \
                                                                               \
  if (tlv_find(cursor, _tlv_tag))                                              \
    goto default_value;                                                        \
                                                                               \
  cfg_struct = (struct _config_struct*)cursor->current;                        \
  *res = _val_type_to_host(cfg_struct->_struct_member);                        \
  return;                                                                      \
                                                                               \
default_value:                                                                 \
  *res = default_val;                                                          \
}                                                                              \
                                                                               \
int gpxe_dyncfg_set_##_opt_name(                                               \
    unsigned char *dyncfg_buf, int dyncfg_buf_size,                            \
    struct sfgpxe_dev_info *dev_info __maybe_unused,                           \
    _val_type val)                                                             \
{                                                                              \
  nvram_partition_t partition;                                                 \
  tlv_cursor_t *cursor = &partition.tlv_cursor;                                \
  struct _config_struct cfg_struct;                                            \
  void *cfg_struct_data = &(cfg_struct._config_struct_first_data_member);      \
  size_t cfg_struct_data_size = sizeof(cfg_struct) - TLV_ITEM_NON_DATA_SIZE;   \
  int rc;								       \
                                                                               \
  /* The default values for the structure are all 0xFF.                        \
   * See tools/mc-comms/ef10config.c:handle_per_pf_string()                    \
   */                                                                          \
  memset(&cfg_struct, -1, sizeof(cfg_struct));                                 \
                                                                               \
  if (tlv_init_partition_from_buffer(&partition,                               \
                                     dyncfg_buf, dyncfg_buf_size)) {           \
    printf("The provided buffer does not represent a valid TLV partition.\n"); \
    return -EINVAL;                                                            \
  }                                                                            \
                                                                               \
  if (TLV_OK == tlv_find(&partition.tlv_cursor, _tlv_tag)) {                   \
    memcpy(&cfg_struct,                                                        \
           (struct _config_struct*)cursor->current,                            \
           sizeof(cfg_struct));                                                \
  }                                                                            \
                                                                               \
  cfg_struct._struct_member = _host_to_val_type(val);                          \
                                                                               \
  rc = tlv_update_or_append_to_partition(	                               \
      &partition,                                                              \
      _tlv_tag,                                                                \
      (const uint8_t*)cfg_struct_data, cfg_struct_data_size);                  \
  if (rc)								       \
    DBGC(dyncfg_buf, "%s: TLV update failed with %x\n", __func__, rc);	       \
  return rc;								       \
}

DEFINE_CONFIG_ACCESSORS(TLV_TAG_FIRMWARE_OPTIONS,
                        tlv_firmware_options, firmware_variant,
                        firmware_variant, unsigned, firmware_variant,
                        host_to_le32, le32_to_host)
DEFINE_CONFIG_ACCESSORS(TLV_TAG_GLOBAL_PCIE_CONFIG,
			tlv_pcie_config, max_pf_number,
			pf_count, int16_t, max_pf_number,
			host_to_le16, le16_to_host)
DEFINE_CONFIG_ACCESSORS(TLV_TAG_GLOBAL_PORT_MODE,
			tlv_global_port_mode, port_mode,
			port_mode, unsigned, port_mode,
			host_to_le32, le32_to_host)
DEFINE_CONFIG_ACCESSORS(TLV_TAG_PRIVILEGE_MASK,
			tlv_privilege_mask, privilege_mask,
			privilege_mask, unsigned, privilege_mask,
			host_to_le32, le32_to_host)
DEFINE_CONFIG_ACCESSORS(TLV_TAG_PFIOV(dev_info->port), tlv_pfiov, pfiov, pfiov,
			uint8_t, pfiov, , )
DEFINE_CONFIG_ACCESSORS(TLV_TAG_VPORT_VLAN_TAG(dev_info->pf_index),
			tlv_vport_vlan_tag, vlan_tag, vlan_tag, uint32_t, vlan_tag,
			host_to_le32, le32_to_host)
DEFINE_CONFIG_ACCESSORS(TLV_TAG_VSWITCH_TYPE(dev_info->port), tlv_vswitch_type,
			vswitch_type, vswitch_type, uint32_t, vswitch_type,
			host_to_le32, le32_to_host)
DEFINE_CONFIG_ACCESSORS(TLV_TAG_GLOBAL_PCIE_CONFIG,
			tlv_pcie_config, max_pf_number,
			vf_aper, uint16_t, vf_aper,
			host_to_le16, le16_to_host)
DEFINE_CONFIG_ACCESSORS(TLV_TAG_OCSD, tlv_ocsd, mode,
			ocsd_mode, uint32_t, mode,
			host_to_le32, le32_to_host)

