/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#ifdef __i386__
/* This is used by the OCSD code.
 * We only care about OCSD for our bootrom, which builds with the i386 arch.
 * There is no such file for the x86_64 arch, which would break compilation.
 */

#include "sfc_hp_ocsd.h"
#include <mc_driver_pcol.h>
#include <realmode.h>
#include <gpxe/instance.h>

int hp_ocsd_start ( struct pci_device *pci )
{
	int rc, status;
	struct smbios_structure ocsd_info;
	size_t size, base_offset, offset;
	struct reserved_mem_entry mem_entry;
	struct ocsd_buf ocsd;
	struct card_entry entry;
	uint32_t ocsd_addr_high, ocsd_addr_low, entry_addr;
	uint32_t temp;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_HP_IN_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_HP_OUT_LEN);
	struct efx_nic *efx = netdev_priv ( pci_get_drvdata ( pci ));
	uint32_t pci_busdevfn_from_prefix = instance_pci_busdevfn();

	uint8_t our_bus = bus_from_busdevfn ( pci_busdevfn_from_prefix );
	uint8_t our_dev = dev_from_busdevfn ( pci_busdevfn_from_prefix );
	uint8_t our_fn  = fn_from_busdevfn ( pci_busdevfn_from_prefix );

	/* Only register OCSD info via the bootrom for function 0. */
	if ( fn_from_busdevfn(pci->devfn) != 0 ||
	     fn_from_busdevfn(pci->devfn) != our_fn ||
	     dev_from_busdevfn(pci->devfn) != our_dev ||
	     pci->bus != our_bus) {
		return 1;
	}


	MCDI_SET_DWORD( inbuf, HP_IN_SUBCMD, MC_CMD_HP_IN_OCSD_SUBCMD );
	MCDI_SET_DWORD( inbuf, HP_IN_OCSD_ADDR_LO, 0 );
	MCDI_SET_DWORD( inbuf, HP_IN_OCSD_ADDR_HI, 0 );
	MCDI_SET_DWORD( inbuf, HP_IN_OCSD_INTERVAL, 1 );
	rc = efx->type->mcdi_rpc( efx, MC_CMD_HP,
				  inbuf, MC_CMD_HP_IN_LEN,
				  outbuf, sizeof( outbuf ),
				  NULL, false);
	status = MCDI_DWORD( outbuf, HP_OUT_OCSD_STATUS );
	if ( rc ) {
		DBG ( "MC firmware, platform, and/or NIC don't support OCSD.\n" );
		goto fail;
	}
	if ( status == MC_CMD_HP_OUT_OCSD_STARTED ||
	     status == MC_CMD_HP_OUT_OCSD_ALREADY_STARTED )
		return 0;

	/* Find the HP SMBIOS structure listing reserved memory entries and find
	 * the entry for OCSD.
	 */
	rc = find_smbios_structure ( smbios_hp_ocsd_type, &ocsd_info );
	if ( rc ) {
		DBG ( "Could not find HP reserved memory smbios structure.\n" );
		goto fail;
	}

	size = ocsd_info.header.len - sizeof ( struct smbios_header );
	if ( size % sizeof ( struct reserved_mem_entry ) != 0 ) {
		DBG ( "Unexpected size for reserved memory entries.\n" );
		goto fail;
	}
	base_offset = ocsd_info.offset + sizeof ( struct smbios_header );
	for (offset = 0;
	     offset < size;
	     offset += sizeof ( mem_entry ) ) {
		copy_smbios_data ( &mem_entry,
				   base_offset + offset,
				   sizeof ( mem_entry ) );
		if ( mem_entry.signature == ocsd_signature )
			break;
	}

	if ( mem_entry.signature != ocsd_signature ) {
		DBG ( "No OCSD entry found.\n" );
		goto fail;
	}

	ocsd_addr_high = (uint32_t)(mem_entry.physical_address >> 32);
	if (ocsd_addr_high != 0) {
		DBG( "Unexpected non-32bit OCSD address.\n" );
		goto fail;
	}
	ocsd_addr_low  = (uint32_t)(mem_entry.physical_address & 0xFFFFFFFFU);
	copy_from_real ( &ocsd,
			 ocsd_addr_high,
			 ocsd_addr_low,
			 sizeof ( ocsd ) );

	if (ocsd.version_number > ocsd_struct_version) {
		DBG( "Unsupported OCSD version %d (supported version: %d)\n",
		     ocsd.version_number,
		     ocsd_struct_version );
		goto fail;
	}

	if (ocsd.card_count >= ocsd.max_card_count) {
		DBG( "Can't register: Max number of cards reached.\n");
		goto fail;
	}

	/* Locally update the fields for the entry. */
	memset( &entry, 0, sizeof ( entry ) );

	entry.card_fields.version_number = option_card_field_version;
	entry.card_fields.bus = bus_from_busdevfn ( pci_busdevfn_from_prefix );
	entry.card_fields.device = dev_from_busdevfn ( pci_busdevfn_from_prefix );
	// TODO(ajr): What is this throttle parameter?
	entry.card_fields.flags = 0;
	/* We don't use the vendor field. */
	entry.card_fields.vendor_defined = 0;
	/* Compute the checksum. */
	entry.card_fields.checksum = 0;
	for ( temp = 0;
	      temp < sizeof (entry.card_fields) / sizeof (uint32_t) - 1;
	      temp++) {
		entry.card_fields.checksum -=
			( ( uint32_t * ) &entry.card_fields )[temp];
	}

	/* Push the card entry. */
	entry_addr = ocsd.buffer_start + ocsd.card_count * ocsd.entry_size;
	copy_to_real ( 0, entry_addr, &entry, sizeof ( entry ) );

	/* Send the MCDI command to tell the firmware to start reporting temperatures. */
	MCDI_SET_DWORD( inbuf, HP_IN_SUBCMD, MC_CMD_HP_IN_OCSD_SUBCMD );
	MCDI_SET_DWORD( inbuf, HP_IN_OCSD_ADDR_HI, 0 );
	MCDI_SET_DWORD( inbuf, HP_IN_OCSD_ADDR_LO,
			 entry_addr + sizeof ( entry.card_fields ) );
	MCDI_SET_DWORD( inbuf, HP_IN_OCSD_INTERVAL, ocsd.requested_update_interval );
	rc = efx->type->mcdi_rpc( efx, MC_CMD_HP,
				  inbuf, MC_CMD_HP_IN_LEN,
				  outbuf, sizeof( outbuf ),
				  NULL, false);
	if ( rc )
		goto fail;

	/* We must update the count only once we are done.
	 * Also don't forget to update the checksum.
	 */
	ocsd.card_count++;
	ocsd.checksum -= 1;
	copy_to_real ( ocsd_addr_high,
		       ocsd_addr_low,
		       &ocsd,
		       sizeof ( ocsd ) );
	MCDI_SET_DWORD ( inbuf, HP_IN_SUBCMD, MC_CMD_HP_IN_OCSD_SUBCMD );
	MCDI_SET_DWORD ( inbuf, HP_IN_OCSD_ADDR_LO, 0 );
	MCDI_SET_DWORD ( inbuf, HP_IN_OCSD_ADDR_HI, 0 );
	MCDI_SET_DWORD ( inbuf, HP_IN_OCSD_INTERVAL, 1 );
	rc = efx->type->mcdi_rpc( efx, MC_CMD_HP,
				  inbuf, MC_CMD_HP_IN_LEN,
				  outbuf, sizeof( outbuf ),
				  NULL, false);
	status = MCDI_DWORD ( outbuf, HP_OUT_OCSD_STATUS );
	if ( rc || status != MC_CMD_HP_OUT_OCSD_STARTED )
		goto fail;

	DBG( "HP OCSD started.\n" );
	return 0;

fail:
	DBG( "Failed to start HP OCSD.\n" );
	return 1;
}

#endif
