/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <gpxe/in.h>
#include <gpxe/ip.h>
#include <gpxe/icmp4.h>
#include <gpxe/iobuf.h>
#include <gpxe/tcp.h>
#include <gpxe/tcpip.h>

struct tcpip_protocol icmp_protocol;

static int icmp_echo_reply ( struct sockaddr_tcpip *st_dest,
			     uint16_t identifier, uint16_t seq, uint32_t data,
			     char *payload, size_t payload_len )
{
	struct io_buffer *iob;
	struct icmp_echo *reply;
	
	iob = alloc_iob ( sizeof ( *reply ) + payload_len + MIN_IOB_LEN );
	if ( NULL == iob )
		return -ENOMEM;
	iob_reserve ( iob, MAX_HDR_LEN + sizeof ( *reply ) );

	/* Copy payload entirely */
	memcpy ( iob_put ( iob, payload_len ), payload, payload_len );

	/* Fill out reply */
	reply = iob_push ( iob, sizeof ( *reply ) );
	reply->type = ICMP_ECHOREPLY;
	reply->code = 0;
	reply->identifier = identifier;
	reply->seq = seq;
	reply->data = data;
	reply->csum = 0;
	reply->csum = tcpip_chksum ( reply, iob_len ( iob ) );

	/* And send the packet */
	return tcpip_tx ( iob, &icmp_protocol, NULL, st_dest, NULL, NULL );
}

static int icmp_echo ( struct io_buffer *iob, struct sockaddr_tcpip *st_src )
{
	struct icmp_echo *echo = iob->data;
	int rc;

	/* Sanity check */
	if ( iob_len ( iob ) < sizeof ( *echo ) ) {
		DBG ( "Packet too short (%zd bytes)\n", iob_len ( iob ) );
		rc = -EINVAL;
		goto out;
	}

	if ( echo->code != 0 ) {
		DBG ( "ICMP ECHO request has non-zero code (%d)\n", echo->code );
		rc = -EINVAL;
		goto out;
	}

	iob_pull ( iob, sizeof ( *echo ) );
	rc = icmp_echo_reply ( st_src, echo->identifier, echo->seq,
			       echo->data, iob->data, iob_len ( iob ) );

out:
	free_iob ( iob );
	return rc;
}

static int icmp_rx ( struct io_buffer *iob, struct sockaddr_tcpip *st_src,
		     __unused struct sockaddr_tcpip *st_dest,
		     __unused uint16_t pshdr_csum )
{
	struct icmp_header *hdr = iob->data;
	uint16_t csum;

	/* Sanity check */
	if ( iob_len ( iob ) < sizeof ( *hdr ) ) {
		DBG ( "Packet too short (%zd bytes)\n", iob_len ( iob ) );
		free_iob ( iob );
		return -EINVAL;
	}

	if ( ( csum = tcpip_chksum ( hdr, iob_len ( iob ) ) ) != 0 ) {
		DBG ( "ICMPv4 checksum incorrect (is %04x including checksum "
		      "field, should be 0000)\n", csum );
		free_iob ( iob );
		return -EINVAL;
	}

	switch ( hdr->type ) {
	case ICMP_ECHO:
		return icmp_echo ( iob, st_src );
	};

	return -ENOSYS;
}

/** ICMP protocol */
struct tcpip_protocol icmp_protocol __tcpip_protocol = {
	.name = "ICMP",
	.rx = icmp_rx,
	.tcpip_proto = IP_ICMP,
};
